/*



Copyright (c) 2001, Granada Learning

All rights reserved.



Redistribution and use in source and binary forms, with or without

modification, are permitted provided that the following conditions are

met:



    * Redistributions of source code must retain the above copyright

	  notice, this list of conditions and the following disclaimer.



    * Redistributions in binary form must reproduce the above copyright

	  notice, this list of conditions and the following disclaimer in the

	  documentation and/or other materials provided with the distribution.



    * Neither the name of the Granada Learning nor the names of its

	  contributors may be used to endorse or promote products derived

	  from this software without specific prior written permission.



THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS

IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,

THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR

PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE

LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR

CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF

SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS

INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN

CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)

ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF

THE POSSIBILITY OF SUCH DAMAGE.



*/



#include "stdafx.h"

#include "WOLauncher.h"





#ifdef _DEBUG

#define new DEBUG_NEW

#undef THIS_FILE

static char THIS_FILE[] = __FILE__;

#endif





using namespace std;





#define APPNAME "WOLauncher"

#define VERSION "1.0"





// Report fatal error and exit

//

void fatalError(LPCSTR message) {

	cerr << endl << "Fatal Error: " << message << endl;

	exit(1);

}







// Gets the pathname of the directory containing the currently executing

// application.

//

void getAppDirectory(CString& strPath)

{ 

	char *buf = strPath.GetBuffer(_MAX_PATH);

	

	// get the filename of the current excutable

	if (GetModuleFileName(NULL, buf, _MAX_PATH)) { 

		// find last directory separator and trucate after it

		char *p = strrchr(buf, '\\');

		if (p) p[1] = 0;

	}



	strPath.ReleaseBuffer();

}





// Normalise directory prefix to Microsoft conventions for consistency (\'s rather that /'s)

// and remove any trailing directory separators

//

void normaliseDirectoryPrefix(CString &prefix) {

	// swap separator style

	prefix.Replace('/', '\\');

	// remove trailing seperator

	if (prefix.Right(1) == "\\") prefix = prefix.Left(prefix.GetLength() - 1);

}









// Gets the NEXT_ROOT directory

//

CString getNextRoot() {

	static CString root;

	static bool    initialised = false;



	if (!initialised) {

		// flag that we've tried

		initialised = true;



		HKEY   hKey;



		// try retrieving the WO base installation directory

		if (RegOpenKeyEx(HKEY_LOCAL_MACHINE,

						 "SOFTWARE\\Apple\\WebObjects",

						 0, KEY_QUERY_VALUE, &hKey) == ERROR_SUCCESS) 

		{

	  		DWORD  bufLen = _MAX_PATH;

			RegQueryValueEx(hKey, "NEXT_ROOT", NULL, NULL,

							(LPBYTE) root.GetBuffer(bufLen), &bufLen);

            RegCloseKey(hKey);



			root.ReleaseBuffer();



			normaliseDirectoryPrefix(root);

		}

		

		if (!root.IsEmpty()) return root;

		cerr << "Warning: Unable to locate WOROOT/NET_ROOT using registry" << endl;





		//

		// ok, can't find it in the registry so hunt for the apple directory

		// in the local drives

		//



		// get a bit mask enumerating the local drives

		DWORD   drive_mask = GetLogicalDrives();

		char    letter = 'A';

		CString drive  = "A:\\";





		while (drive_mask != 0) {



			// get the drive root

			drive.SetAt(0, letter++);



			// skip missing letters

			boolean valid_drive = drive_mask & 1;

			drive_mask >>= 1;

			if (!valid_drive) continue;



			// find out the type

			UINT drive_type = GetDriveType(drive);



			// skip any drive that isn't fixed or remote (cdroms, removable storage, etc)

			if (drive_type != DRIVE_FIXED && drive_type != DRIVE_REMOTE) continue;



			

			// try looking for the Apple root directory on this drive

			root = drive + "Apple";



			cerr << "Looking for WOROOT in " << ((LPCSTR) root) << endl;





			CFileStatus status;

			if (CFile::GetStatus(root, status)) {

				// find a file with the right name - is it a directory?

				if (status.m_attribute == 0x10) {

					// yes!

					return root;

				}

				

			}



		}



		fatalError("Unable to locate WOROOT/NET_ROOT - is WebObjects installed?");



	}



	return root;

}





// -- Application Environment --------------------------------------------



// Path to Current directory

CString g_currentDirectory;



// Path to the application Contents directory

CString g_appRoot;



// Path to WebObjects installation

CString g_WORoot(getNextRoot());



// Path to WebObjects Local directory

CString g_localRoot(getNextRoot() + "\\Local");



// Path to Home Root (what does this mean in this context?)

CString g_homeRoot("C:\\");



// Application Class

CString g_appClass("Application");



// JVM command

CString g_jvmCommand("java");



// JVM options

CString g_jvmOptions("");





// -- Launch Property Handling -------------------------------------------



// Set launch options specified in the classpath file

//

void setLaunchProperty(CString key, CString value) {

	if (key == "JVM") {

		g_jvmCommand = value;

	} else if (key == "JVMOptions") {

		g_jvmOptions = value;

	} else if (key == "ApplicationClass") {

		g_appClass = value;

	}

}





// -- Path Substitution --------------------------------------------------





// Replaces prefixFrom with prefixTo if found at the start of the supplied

// path.

//

void expandPath(CString& path, LPCSTR prefixFrom, LPCSTR prefixTo) {

	int len = strlen(prefixFrom);



	if (path.Left(len) == prefixFrom) {

		CString fixed(prefixTo);

		path = prefixTo + path.Mid(len);

	}

}



// Expand the standard prefixes at the start of the supplied path.

//

void expandPath(CString& path) {

	expandPath(path, "WOROOT",    g_WORoot);

	expandPath(path, "LOCALROOT", g_localRoot);

	expandPath(path, "APPROOT",   g_appRoot);

	expandPath(path, "HOMEROOT",  g_homeRoot);

}





// -- Launch command creation --------------------------------------------





// Build the classpath from the platform classpath file

//

void initialiseClassPath(CString& classpath)

{ 

	CStdioFile	in(g_appRoot + "\\Windows\\CLSSPATH.TXT", CFile::modeRead);

	CString line;



	while (in.ReadString(line)) {

		line.TrimLeft();

		line.TrimRight();



		// is it a comment?

		if (line.GetAt(0) == '#') {

			// yes - ok, does it define a launcher property?

			int i = line.Find(" == ");

			if (i > -1) {

				CString key = line.Mid(1, i - 1);

				key.TrimLeft();

				key.TrimRight();

				

				CString value = line.Mid(i + 4);

				value.TrimLeft();

				value.TrimRight();

			   			   

				setLaunchProperty(key, value);

			}



			continue;

		}



		// Expand the path prefixes

		expandPath(line);



		if (!classpath.IsEmpty()) classpath += ";";

		classpath += line;

	}





	

}







// Get the property declarations for the various WO Paths

//

CString getDirectoryProperties() {

	CString args;



	

	args += "\"-DWORootDirectory=";

	args += g_WORoot;

	args += "\" ";





	args += "\"-DWOLocalRootDirectory=";

	args += g_localRoot;

	args += "\" ";



	args += "\"-DWOUserDirectory=";

	args += g_currentDirectory;

	args += "\" ";

	



	return args;

}





//  Build the command to executed from it's component parts

//

void buildCommand(CString& cmd, LPCSTR classpath, int argc, TCHAR* argv[]) {





	CString jvmArgs = getDirectoryProperties() + g_jvmOptions;

	CString appArgs;

	

	// process on the args - skipping arg 0 (command name)

	for (int i = 1; i < argc; i++) {

		TCHAR *arg = argv[i];



		// special handing jvm arguments

		if (arg[0] == '-' && (			

			   (arg[1] == 'X') ||                      // jvm "special" arguments

			   (arg[1] == 'D' && strchr(arg + 2, '=')) // property definitions

			))

		{

				jvmArgs += " ";

				jvmArgs += arg;

				continue;

		}

		

		

		// otherwise append it to the command string

		appArgs += " ";

		appArgs += argv[i];

	}



	// build the command

	cmd.Format("%s %s -cp \"%s\" %s%s",

			   (LPCSTR)g_jvmCommand, (LPCSTR)jvmArgs,

			   classpath, (LPCSTR)g_appClass, (LPCSTR)appArgs);

	

}









int _tmain(int argc, TCHAR* argv[], TCHAR* envp[])

{

	// initialize MFC and print and error on failure

	if (!AfxWinInit(::GetModuleHandle(NULL), NULL, ::GetCommandLine(), 0))

	{

		cerr << "Fatal Error: MFC initialization failed" << endl;



		// bug out here

		return 1;

	}



	try {

		// tell the world about ourselves;

		cout << APPNAME << " " << VERSION << endl;



		// where are we being launched from?

		getAppDirectory(g_currentDirectory);

		normaliseDirectoryPrefix(g_currentDirectory);

		g_appRoot = g_currentDirectory + "\\Contents";



		// get the classpath

		CString classpath;

		initialiseClassPath(classpath);

		



		// make the command line

		CString cmd;

		buildCommand(cmd, classpath, argc, argv);



		// tell the user

		cout << "Executing command: " << endl << (LPCTSTR) cmd << endl;



		// set the current directory

		SetCurrentDirectory(g_currentDirectory);



		// begin next lines added/edited by Codeon GmbH, 21.05.2002

		// go!

		// system (cmd);



		PROCESS_INFORMATION processInformation;

		STARTUPINFO startupInfo;

		char *commandLine;

 

		// prepare CreateProcess call

		::memset(&processInformation, 0, sizeof(PROCESS_INFORMATION));

		::memset(&startupInfo, 0, sizeof(STARTUPINFO));



		// CreateProcess maybe alter the commandline, so grant it

		commandLine = new char[cmd.GetLength() << 1];

		if (commandLine)

			::strcpy(commandLine, (LPCSTR)cmd);



		// create a new child process

		if (commandLine && ::CreateProcess (NULL, commandLine, NULL, NULL, TRUE, 0, 

										    NULL, NULL, &startupInfo, &processInformation)) 

		{

			// wait until child process terminates

			::WaitForMultipleObjects(1, &processInformation.hProcess, FALSE, INFINITE);



			::CloseHandle(processInformation.hProcess);

		}

		else

			cerr << "WOLancher: CreateProcess failed!" << endl;



		delete commandLine;



		// end next lines added/edited by Codeon GmbH, 21.05.2002

	} 

	catch (CException *e) 

	{

		// report to any logging

		char msg[1024];

		e->GetErrorMessage(msg, sizeof(msg));

		cerr << "WOLancher exception thrown:\n" << msg;



		// report to as a popup error

		e->ReportError();



		// delete the exception - MFC truly sucks

		e->Delete();

		return 1; // failure

	}



	return 0; // ok!

}















